
from collections import OrderedDict
import copy
import itertools
import os
import yaml

from . import instances
from . import util

class MatrixScope:
	@staticmethod
	def walk_matrix(cfg, root_yml, expand):
		outmost = MatrixScope(cfg, None)

		result = set()
		if 'matrix' not in root_yml:
			result.update(expand(outmost))
		else:
			for item_yml in root_yml['matrix']['include']:
				scope = MatrixScope(cfg, outmost)
				scope.select(item_yml)
				result.update(expand(scope))

		return result

	def __init__(self, cfg, outer):
		self.cfg = cfg

		if outer is not None:
			self.selected_exps = copy.copy(outer.selected_exps)
			self.selected_revs = copy.copy(outer.selected_revs)
			self.selected_axes = copy.copy(outer.selected_axes)
			self.selected_variants = copy.deepcopy(outer.selected_variants)
			self.selected_instsets = copy.copy(outer.selected_instsets)
			self.num_repetitions = outer.num_repetitions
		else:
			self.selected_exps = None
			self.selected_revs = None
			self.selected_axes = None
			self.selected_variants = {}
			self.selected_instsets = None
			self.num_repetitions = None

	def select(self, item_yml):
		if 'experiments' in item_yml:
			if self.selected_exps is None:
				self.selected_exps = set()
			self.selected_exps.update(item_yml['experiments'])

		if 'revisions' in item_yml:
			if self.selected_revs is None:
				self.selected_revs = set()
			self.selected_revs.update(item_yml['revisions'])

		if 'axes' in item_yml:
			if self.selected_axes is None:
				self.selected_axes = set()
			self.selected_axes.update(item_yml['axes'])

		if 'variants' in item_yml:
			for name in item_yml['variants']:
				variant = self.cfg.get_variant(name)
				if variant.axis not in self.selected_variants:
					self.selected_variants[variant.axis] = set()
				self.selected_variants[variant.axis].add(variant.name)

		if 'instsets' in item_yml:
			if self.selected_instsets is None:
				self.selected_instsets = set()
			self.selected_instsets.update(item_yml['instsets'])

		if 'repeat' in item_yml:
			self.num_repetitions = item_yml['repeat']

class Config:
	"""Represents the entire configuration (i.e., an experiments.yml file)."""

	def __init__(self, basedir, yml):
		assert os.path.isabs(basedir)
		self.basedir = basedir
		self.yml = yml

		self._insts = OrderedDict()
		self._build_infos = OrderedDict()
		self._revisions = OrderedDict()
		self._variants = OrderedDict()
		self._exp_infos = OrderedDict()

		def construct_instances():
			if 'instances' in self.yml:
				for inst_yml in self.yml['instances']:
					for item in inst_yml['items']:
						yield Instance(self, item, inst_yml)

		def construct_variants():
			if 'variants' in self.yml:
				for axis_yml in self.yml['variants']:
					for variant_yml in axis_yml['items']:
						yield Variant(self, axis_yml['axis'], variant_yml)

		for inst in sorted(construct_instances(), key=lambda inst: inst.filename):
			self._insts[inst.filename] = inst

		if 'builds' in self.yml:
			for build_yml in sorted(self.yml['builds'], key=lambda y: y['name']):
				self._build_infos[build_yml['name']] = BuildInfo(self, build_yml)

		if 'revisions' in self.yml:
			for revision_yml in sorted(self.yml['revisions'], key=lambda y: y['name']):
				self._revisions[revision_yml['name']] = Revision(self, revision_yml)

		for variant in sorted(construct_variants(), key=lambda variant: variant.name):
			self._variants[variant.name] = variant

		if 'experiments' in self.yml:
			for exp_yml in sorted(self.yml['experiments'], key=lambda y: y['name']):
				self._exp_infos[exp_yml['name']] = ExperimentInfo(self, exp_yml)

	def instance_dir(self):
		"""Path of the directory that stores all the instances."""
		return os.path.join(self.basedir, self.yml['instdir'])

	def all_instance_ids(self):
		for inst in self.all_instances():
			yield inst.filename

	def all_instances(self):
		yield from self._insts.values()

	def get_instance(self, name):
		if name not in self._insts:
			raise RuntimeError("Instance {} does not exist".format(name))
		return self._insts[name]

	def get_build_info(self, name):
		if name not in self._build_infos:
			raise RuntimeError("BuildInfo {} does not exist".format(name))
		return self._build_infos[name]

	def all_revisions(self):
		yield from self._revisions.values()

	def get_revision(self, name):
		if name is None: # TODO: Questionable special case.
			return None
		if name not in self._revisions:
			raise RuntimeError("Revision {} does not exist".format(name))
		return self._revisions[name]

	def all_builds(self):
		if 'builds' in self.yml:
			for build_yml in sorted(self.yml['builds'], key=lambda y: y['name']):
				for revision in self.all_revisions():
					spec_set = set(revision.specified_versions)
					if build_yml['name'] not in spec_set:
						continue
					# TODO: Exclude the build if not all requirements are specified in spec_set.
					yield Build(self, self.get_build_info(build_yml['name']), revision)

	def all_builds_for_revision(self, revision):
		for build in self.all_builds():
			if build.revision == revision:
				yield build

	def get_build(self, name, revision):
		for build in self.all_builds(): # TODO: Avoid a quadratic blowup.
			if build.name == name and build.revision == revision:
				return build
		raise RuntimeError("Build {} does not exist".format(name))

	def all_variants(self):
		yield from self._variants.values()

	def get_variant(self, name):
		if name not in self._variants:
			raise RuntimeError("Variant {} does not exist".format(name))
		return self._variants[name]

	def _test_variation_id_in_scope(self, variation_id, scope):
		if scope.selected_axes is not None:
			for name in variation_id:
				variant = self.get_variant(name)
				if variant.axes not in scope.selected_axes:
					return False
		if scope.selected_variants is not None:
			for name in variation_id:
				variant = self.get_variant(name)
				if variant.axis in scope.selected_variants:
					if variant.name not in scope.selected_variants[variant.axis]:
						return False
		return True

	# Determine all variations selected by a scope.
	def _expand_variation_ids_in_scope(self, scope):
		if scope.selected_axes is None:
			axes = {var.axis for var in self.all_variants()}
		else:
			axes = scope.selected_axes

		variants = {}
		for axis in axes:
			if scope.selected_variants is None or axis not in scope.selected_variants:
				variants[axis] = {var.name for var in self.all_variants() if var.axis == axis}
			else:
				variants[axis] = scope.selected_variants[axis]

		variation_bundle = [ ]
		for axis_variants in variants.values():
			# Sort once so that variation order is deterministic.
			variant_list = sorted(axis_variants, key=lambda name: name if name is not None else '')
			variation_bundle.append(variant_list)

		# A variation is defined as a tuple of variants.
		def make_variation(prod):
			variant_filter = filter(lambda name: name is not None, prod)
			# Sort again so that order of the variants does not depend on the axes.
			variant_list = sorted(variant_filter)
			return tuple(variant_list)

		return [make_variation(prod) for prod in itertools.product(*variation_bundle)]

	def get_experiment_info(self, name):
		if name not in self._exp_infos:
			raise RuntimeError("Experiment {} does not exist".format(name))
		return self._exp_infos[name]

	def all_experiments(self):
		for exp_name, rev_name, var_names in self._expand_experiment_matrix():
			revision = self.get_revision(rev_name)
			variation = [self.get_variant(var_name) for var_name in var_names]
			yield Experiment(self, self.get_experiment_info(exp_name), revision, variation)

	def _expand_experiment_matrix(self):
		def expand(scope):
			# Determine all experiments selected by this scope.
			if scope.selected_exps is None:
				experiments = [exp_yml['name'] for exp_yml in self.yml['experiments']]
			else:
				experiments = scope.selected_exps

			# Helper to find all selected revisions for a given experiment.
			def revisions_for_experiment(exp_name):
				if scope.selected_revs is None:
					if 'use_builds' in self.get_experiment_info(exp_name)._exp_yml:
						for revision in self.all_revisions():
							yield revision.name
					else:
						yield None
				else:
					yield from scope.selected_revs

			variation_ids = self._expand_variation_ids_in_scope(scope)
			for exp_name in experiments:
				yield from itertools.product([exp_name], revisions_for_experiment(exp_name),
						variation_ids)

		expansion = MatrixScope.walk_matrix(self, self.yml, expand)
		return sorted(expansion)

	def _experiment_matches_item(self, item_yml, name, revision):
		if 'experiments' in item_yml:
			if name not in item_yml['experiments']:
				return False
		if revision is None:
			if 'revisions' in item_yml:
				return False
		else:
			if 'revisions' in item_yml:
				if revision.name not in item_yml['revisions']:
					return False
		return True

	def discover_all_runs(self):
		for exp in self.all_experiments():
			for inst_name, rep in self._expand_run_matrix(exp):
				instance = self.get_instance(inst_name)
				yield Run(self, exp, instance, rep)

	def _expand_run_matrix(self, exp):
		def expand(scope):
			if scope.selected_exps is not None:
				if exp.name not in scope.selected_exps:
					return
			if scope.selected_revs is not None:
				if exp.revision.name not in scope.selected_revs:
					return
			if scope.selected_variants is not None:
				variation_id = tuple(var.name for var in exp.variation)
				if variation_id not in self._expand_variation_ids_in_scope(scope):
					return

			if scope.selected_instsets is not None:
				instsets = scope.selected_instsets
			else:
				instset_combinations = [instance.instsets for instance in self.all_instances()]
				instsets = set().union(*instset_combinations)

			for instance in self.all_instances():
				if instsets.isdisjoint(instance.instsets):
					continue

				reps = range(0, 1)
				if scope.num_repetitions is not None:
					reps = range(0, scope.num_repetitions)
				elif 'repeat' in exp.info._exp_yml:
					reps = range(0, exp.info._exp_yml['repeat'])
				for rep in reps:
					yield (instance.filename, rep)

		expansion = MatrixScope.walk_matrix(self, self.yml, expand)
		return sorted(expansion)

	def collect_successful_results(self, parse_fn):
		"""
		Collects all success runs and parses their output.

		:param: parse_fn: Function to parse the output. Takes two parameters
			(run, f) where run is a :class:`simexpal.base.Run` object and f
			is a Python file object.
		"""

		res = [ ]
		for run in self.discover_all_runs():
			exp = run.experiment
			finished = os.access(run.output_file_path('status'), os.F_OK)
			if not finished:
				print("Skipping unfinished run {}/{}[{}]".format(run.experiment.name,
						run.instance.shortname, run.repetition))
				continue

			with open(run.output_file_path('status'), "r") as f:
				status_dict = yaml.load(f, Loader=yaml.Loader)
			if status_dict['timeout'] or status_dict['signal'] or status_dict['status'] > 0:
				print("Skipping failed run {}/{}[{}]".format(run.experiment.name,
						run.instance.shortname, run.repetition))
				continue

			with open(run.output_file_path('out'), 'r') as f:
				res.append(parse_fn(run, f))
		return res

class Instance:
	"""Represents a single instance"""

	def __init__(self, cfg, filename, inst_yml):
		self._cfg = cfg
		self.filename = filename
		self._inst_yml = inst_yml

	@property
	def config(self):
		return self._cfg

	@property
	def shortname(self):
		return os.path.splitext(self.filename)[0]

	@property
	def fullpath(self):
		return os.path.join(self._cfg.instance_dir(), self.filename)

	@property
	def instsets(self):
		if 'set' not in self._inst_yml:
			return set([None])
		if isinstance(self._inst_yml['set'], list):
			return set(self._inst_yml['set'])
		assert isinstance(self._inst_yml['set'], str)
		return set([self._inst_yml['set']])

	@property
	def repo(self):
		if 'repo' not in self._inst_yml:
			return None
		return self._inst_yml['repo']

	def check_available(self):
		return os.path.isfile(os.path.join(self._cfg.instance_dir(), self.filename))

	def install(self):
		if self.check_available():
			return

		util.try_mkdir(self._cfg.instance_dir())

		if 'repo' in self._inst_yml:
			if self._inst_yml['repo'] == 'local':
				return

		partial_path = os.path.join(self._cfg.instance_dir(), self.filename)
		if 'repo' in self._inst_yml:
			print("Downloading instance '{}' from {} repository".format(self.filename,
					self._inst_yml['repo']))

			instances.download_instance(self._inst_yml,
					self.config.instance_dir(), self.filename, partial_path, '.post0')
		else:
			assert 'generator' in self._inst_yml
			import subprocess

			def substitute(p):
				if p == 'INSTANCE_FILENAME':
					return self.filename
				raise RuntimeError("Unexpected parameter {}".format(p))

			print("Generating instance '{}'".format(self.filename))

			assert isinstance(self._inst_yml['generator']['args'], list)
			cmd = [util.expand_at_params(arg_tmpl, substitute) for arg_tmpl
					in self._inst_yml['generator']['args']]

			with open(partial_path + '.gen', 'w') as f:
				subprocess.check_call(cmd, cwd=self.config.basedir,
						stdout=f, stderr=subprocess.PIPE)
			os.rename(partial_path + '.gen', partial_path + '.post0')

		stage = 0
		if 'postprocess' in self._inst_yml:
			assert self._inst_yml['postprocess'] == 'to_edgelist'
			instances.convert_to_edgelist(self._inst_yml,
					partial_path + '.post0', partial_path + '.post1');
			os.unlink(partial_path + '.post0')
			stage = 1

		os.rename(partial_path + '.post{}'.format(stage), partial_path)

	def run_transform(self, transform, out_path):
		stage = 0
		assert transform == 'to_edgelist'
		instances.convert_to_edgelist(self._inst_yml,
				self.fullpath, out_path + '.transf1');
		stage = 1

		os.rename(out_path + '.transf{}'.format(stage), out_path)

class BuildInfo:
	def __init__(self, cfg, build_yml):
		self._cfg = cfg
		self._build_yml = build_yml

	@property
	def name(self):
		return self._build_yml['name']

	@property
	def requirements(self):
		if 'requires' in self._build_yml:
			for name in self._build_yml['requires']:
				yield name

	def traverse_requirements(self):
		# Perform a DFS to discover all recursively required builds.
		stack = []
		visited = set()

		for req_name in self.requirements:
			assert req_name not in visited
			req_info = self._cfg.get_build_info(req_name)
			stack.append(req_info)
			visited.add(req_name)

		while len(stack):
			current = stack.pop()
			yield current
			for req_name in current.requirements:
				if req_name in visited:
					continue
				req_info = self._cfg.get_build_info(req_name)
				stack.append(req_info)
				visited.add(req_name)

class Revision:
	def __init__(self, cfg, revision_yml):
		self._cfg = cfg
		self.revision_yml = revision_yml

	@property
	def name(self):
		return self.revision_yml['name']

	@property
	def specified_versions(self):
		return self.revision_yml['build_version'].keys()

	def version_for_build(self, build_name):
		return self.revision_yml['build_version'][build_name]

class Build:
	def __init__(self, cfg, info, revision):
		self._cfg = cfg
		self.info = info
		self.revision = revision

	@property
	def name(self):
		return self.info.name

	@property
	def clone_dir(self):
		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev + '.clone')

	@property
	def compile_dir(self):
		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev + '.compile')

	@property
	def prefix_dir(self):
		rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'builds', self.name + rev)

class Variant:
	def __init__(self, cfg, axis, variant_yml):
		self._cfg = cfg
		self.axis = axis
		self.variant_yml = variant_yml

	@property
	def name(self):
		return self.variant_yml['name']

class ExperimentInfo:
	def __init__(self, cfg, exp_yml):
		self._cfg = cfg
		self._exp_yml = exp_yml

	@property
	def name(self):
		return self._exp_yml['name']

	@property
	def used_builds(self):
		if 'use_builds' in self._exp_yml:
			for name in self._exp_yml['use_builds']:
				yield name

class Experiment:
	"""
	Represents an experiment (see below).

	An experiment is defined as a combination of command line arguments
	and environment
	(from the experiment stanza in a experiments.yml file),
	a revision that is used to build the experiment's program
	and a set of variants (from the variants stanza in a experiments.yml file).
	"""

	def __init__(self, cfg, info, revision, variation):
		self._cfg = cfg
		self.info = info
		self.revision = revision
		self.variation = variation

	@property
	def name(self):
		return self.info.name

	@property
	def aux_subdir(self):
		var = ''
		if self.variation:
			var = '~' + ','.join([variant.name for variant in self.variation])
		rev = ''
		if self.revision:
			rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'aux', self.name + var + rev)

	@property
	def output_subdir(self):
		var = ''
		if self.variation:
			var = '~' + ','.join([variant.name for variant in self.variation])
		rev = ''
		if self.revision:
			rev = '@' + self.revision.name
		return os.path.join(self._cfg.basedir, 'output', self.name + var + rev)

class Run:
	def __init__(self, cfg, experiment, instance, repetition):
		self._cfg = cfg
		self.experiment = experiment
		self.instance = instance
		self.repetition = repetition

	@property
	def config(self):
		return self._cfg

	# Contains auxiliary files that SHOULD NOT be necessary to determine the result of the run.
	def aux_file_path(self, ext):
		(fbase, _) = os.path.splitext(self.instance.filename)
		rep = ''
		if self.repetition > 0:
			rep = '[{}]'.format(self.repetition)
		return os.path.join(self.experiment.aux_subdir, fbase + '.' + ext + rep)

	# Contains the final output files; those SHOULD be all that is necessary to determine
	# if the run succeeded and to evaluate its result.
	def output_file_path(self, ext):
		(fbase, _) = os.path.splitext(self.instance.filename)
		rep = ''
		if self.repetition > 0:
			rep = '[{}]'.format(self.repetition)
		return os.path.join(self.experiment.output_subdir, fbase + '.' + ext + rep)

def read_and_validate_setup(basedir='.', setup_file='experiments.yml'):
	return util.validate_setup_file(os.path.join(basedir, setup_file))

def config_for_dir(basedir=None):
	if basedir is None:
		basedir = '.'
	yml = read_and_validate_setup(basedir=basedir)
	return Config(os.path.abspath(basedir), read_and_validate_setup(basedir=basedir))





import os
import subprocess
import sys
import tempfile
import yaml

from .. import util
from . import common

dry_run = False

class SgeLauncher(common.Launcher):
	def __init__(self, queue):
		self.queue = queue

	def submit(self, run):
		self._do_submit(run.config, True, run)

	def submit_multiple(self, config, runs):
		self._do_submit(config, False, runs)

	def _do_submit(self, config, single_task, r):
		util.try_mkdir(os.path.join(config.basedir, 'aux'))
		util.try_mkdir(os.path.join(config.basedir, 'aux/_sge'))

		script = os.path.abspath(sys.argv[0])

		sge_args = ['qsub', '-b', 'y', '-q', self.queue]
		sge_args.extend(['-o', os.path.join(config.basedir, 'aux/_sge/$JOB_ID.out'),
				'-e', os.path.join(config.basedir, 'aux/_sge/$JOB_ID.err')])
		invoke_args = [script, '-C', config.basedir, 'invoke']

		if single_task:
			if not common.lock_run(r):
				return
			locked = [r]
			print("Launching experiment '{}', instance '{}' on SGE queue '{}'".format(
					r.experiment.name, r.instance.filename, self.queue))

			invoke_args.extend(['--experiment', r.experiment.name,
					'--instance', r.instance.filename,
					'--repetition', str(r.repetition)])
		else:
			locked = [ ]
			for run in r:
				if not common.lock_run(run):
					continue
				print("Launching experiment '{}', instance '{}' on SGE queue '{}'".format(
						run.experiment.name, run.instance.filename, self.queue))
				locked.append(run)

			if not locked:
				return

			spec_dict = {'array': [
				{'experiment': run.experiment.name,
					'instance': run.instance.filename,
					'repetition': run.repetition}
				for run in locked
			]}
			(specfd, specfile) = tempfile.mkstemp(prefix='', suffix='.spec',
					dir=os.path.join(run.config.basedir, 'aux/_sge'))
			with os.fdopen(specfd, 'w') as f:
				yaml.dump(spec_dict, f, default_flow_style=False)

			sge_args.extend(['-t', '{}-{}'.format(0, len(locked)-1)])
			invoke_args.extend(['--specfile', specfile, '--sge-index'])

		if not dry_run:
			subprocess.check_call(sge_args + invoke_args);
		else:
			print("Would invoke SGE as:", sge_args + invoke_args)
			if single_task:
				subprocess.check_call(invoke_args + ['-n'])
			else:
				for i in range(len(locked)):
					sim_env = os.environ.copy()
					sim_env['SGE_TASK_ID'] = str(i)
					subprocess.check_call(invoke_args + ['-n'], env=sim_env)

		for run in locked:
			common.create_run_file(run)




import errno
import gzip
import os
import requests
import shutil
import tarfile
import yaml
import zipfile

from .util import try_mkdir

class DownloadException(Exception):
	pass

repos = {
	'konect': {
		'url': 'http://konect.cc/files/download.tsv.',
		'file_fmt': '.tar.bz2',
	},
	'snap': {
			'url': 'https://snap.stanford.edu/data/',
			'file_fmt': '.txt.gz'
	},
	'network_repository': {
		'url': 'http://nrvis.com/download/data/',
		'file_fmt': '.zip'
	}
}

def download_instance(inst_yml, instances_dir, filename, partial_path, ext):
	repo = inst_yml['repo']

	try_mkdir(instances_dir)

	# Download the instance from the repository.
	prefix = ''
	if 'repo-subdir' in inst_yml:
		prefix = inst_yml['repo-subdir'] + '/'
	fmt = repos[repo]['file_fmt']
	url = repos[repo]['url'] + prefix + filename + fmt

	download_path = os.path.join(instances_dir, filename + '.download')
	request = requests.get(url)
	with open(download_path, 'wb') as f:
		f.write(request.content)

	# Decompress the instance.
	def extract(reader, f):
		f.write(reader.read())

	tmp_path = os.path.join(instances_dir, filename + '.tmp')
	compression = fmt.split('.')[-1]
	if repo == 'konect':
		tar = tarfile.open(download_path, 'r:' + compression)
		member_name = next(elem for elem in tar.getnames() if 'out.' in elem)

		with tar.extractfile(member_name) as reader:
			with open(tmp_path, 'wb') as f:
				extract(reader, f)
	elif repo == 'snap':
		with gzip.open(download_path, 'rb') as reader:
			with open(tmp_path, 'wb') as f:
				extract(reader, f)
	elif repo == 'network_repository':
		zip_file = zipfile.ZipFile(download_path + compression, 'r')
		# TODO finish
	else:
		DownloadException('Unknown repository: ' + repo)

	os.unlink(download_path)
	os.rename(tmp_path, partial_path + ext)

# Reformats the network to a SNAP/EdgeList format.
def convert_to_edgelist(inst_yml, in_path, out_path):
	repo = inst_yml['repo']

	lines = []
	if repo == 'konect':
		separator = ' '
		commentPrefix = '%'
	elif repo == 'snap':
		separator = '\t'
		commentPrefix = '#'

	def get_other_separator(separator):
		if separator == ' ':
			return '\t'
		return ' '
	with open(in_path, 'r') as in_f:
		with open(out_path, 'w') as out_f:
			for line in in_f:
				if line.startswith(commentPrefix):
					continue
				split_line = line.strip().split(separator)
				if len(split_line) < 2:
					separator = get_other_separator(separator)
					split_line = line.strip().split(separator)
					if len(split_line) < 2:
						raise Exception("Unknown separator: " + line.strip())
				split_line = [x for x in filter(lambda x: len(x) > 0, split_line)]
				(u, v) = split_line[0], split_line[1]
				out_f.write("{:s} {:s}\n".format(u, v))





import os
import selectors
import signal
import subprocess
import time

import yaml

from .. import util

class Launcher:
	pass

def lock_run(run):
	(exp, instance) = (run.experiment, run.instance.filename)
	util.try_mkdir(os.path.join(run.config.basedir, 'aux'))
	util.try_mkdir(os.path.join(run.config.basedir, 'output'))
	util.try_mkdir(exp.aux_subdir)
	util.try_mkdir(exp.output_subdir)

	# We will try to launch the experiment.
	# First, create a .lock file. If that is successful, we are the process that
	# gets to launch the experiment. Afterwards, concurrent access to our files
	# can be considered a bug (or deliberate misuse) and will lead to hard failues.
	try:
		lockfd = os.open(run.aux_file_path('lock'),
				os.O_RDONLY | os.O_CREAT | os.O_EXCL, mode=0)
	except FileExistsError:
		# TODO: Those warnings should be behind a flag.
#				print("Warning: .lock file exists for experiment '{}', instance '{}'".format(
#						exp.name, instance))
#				print("Either experiments are launched concurrently or the launcher crashed.")
		return False
	os.close(lockfd)
	return True

def create_run_file(run):
	(exp, instance) = (run.experiment, run.instance.filename)

	# Create the .run file. This signals that the run has been submitted.
	with open(run.aux_file_path('run.tmp'), "w") as f:
		pass
	os.rename(run.aux_file_path('run.tmp'), run.aux_file_path('run'))

def invoke_run(run):
	(exp, instance) = (run.experiment, run.instance.filename)

	# Perform a DFS to discover all used builds.
	recusive_builds = []
	builds_visited = set()

	for name in exp.info.used_builds:
		assert name not in builds_visited
		build = run.config.get_build(name, exp.revision)
		recusive_builds.append(build)
		builds_visited.add(name)

	i = 0 # Need index-based loop as recusive_builds is mutated in the loop.
	while i < len(recusive_builds):
		build = recusive_builds[i]
		for req_name in build.info.requirements:
			if req_name in builds_visited:
				continue
			req_build = run.config.get_build(req_name, exp.revision)
			recusive_builds.append(req_build)
			builds_visited.add(req_name)
		i += 1

	# Collect extra arguments from variants
	extra_args = [ ]
	for variant in exp.variation:
		extra_args.extend(variant.variant_yml['extra_args'])

	# Create the output file. This signals that the run has been started.
	stdout = None
	with open(run.output_file_path('out'), "w") as f:
		# We do not actually need to write anything to the output file.
		# However, we might want to pipe experimental output to it.
		if 'output' in exp.info._exp_yml and exp.info._exp_yml['output'] == 'stdout':
			stdout = os.dup(f.fileno())

	def substitute(p):
		if p == 'INSTANCE':
			return run.config.instance_dir() + '/' + instance
		elif p == 'REPETITION':
			return str(run.repetition)
		elif p == 'OUTPUT':
			return run.output_file_path('out')
		else:
			return None

	def substitute_list(p):
		if p == 'EXTRA_ARGS':
			return extra_args
		else:
			return None

	assert isinstance(exp.info._exp_yml['args'], list)
	cmd = util.expand_at_params(exp.info._exp_yml['args'], substitute, listfn=substitute_list)

	# Build the environment.
	def prepend_env(var, items):
		if(var in os.environ):
			return ':'.join(items) + ':' + os.environ[var]
		return ':'.join(items)

	build_paths = [os.path.join(p.prefix_dir, 'bin') for p in recusive_builds]
	build_ld_paths = [os.path.join(p.prefix_dir, 'lib') for p in recusive_builds]

	environ = os.environ.copy()
	environ['PATH'] = prepend_env('PATH', build_paths)
	environ['LD_LIBRARY_PATH'] = prepend_env('LD_LIBRARY_PATH', build_ld_paths)

	if 'environ' in exp.info._exp_yml:
		for (k, v) in exp.info._exp_yml['environ'].items():
			environ[k] = str(v)

	# Dumps data from an FD to the FS.
	# Creates the output file only if something is written.
	class LazyWriter:
		def __init__(self, fd, path):
			self._fd = fd
			self._path = path
			self._out = None

		def progress(self):
			# Specify some chunk size to avoid reading the whole pipe at once.
			chunk = self._fd.read(16 * 1024)
			if not len(chunk):
				return False

			if self._out is None:
				self._out = open(self._path, "wb")
			self._out.write(chunk)
			return True

		def close(self):
			if self._out is not None:
				self._out.close()

	start = time.perf_counter()
	child = subprocess.Popen(cmd, cwd=run.config.basedir, env=environ,
			stdout=stdout, stderr=subprocess.PIPE)
	sel = selectors.DefaultSelector()
	
	stderr_writer = LazyWriter(child.stderr, run.aux_file_path('stderr'))
	sel.register(child.stderr, selectors.EVENT_READ, stderr_writer)

	# Wait until the run program finishes.
	while True:
		if child.poll() is not None:
			break

		elapsed = time.perf_counter() - start
		if 'timeout' in exp.info._exp_yml and elapsed > float(exp.info._exp_yml['timeout']):
			child.send_signal(signal.SIGXCPU)

		# Consume any output that might be ready.
		events = sel.select(timeout=1)
		for (sk, mask) in events:
			if not sk.data.progress():
				sel.unregister(sk.fd)

	# Consume all remaining output.
	while True:
		events = sel.select(timeout=0)
		for (sk, mask) in events:
			if not sk.data.progress():
				sel.unregister(sk.fd)
		if not events:
			break
	stderr_writer.close()
	runtime = time.perf_counter() - start

	# Collect the status information.
	status = None
	sigcode = None
	if child.returncode < 0: # Killed by a signal?
		sigcode = signal.Signals(-child.returncode).name
	else:
		status = child.returncode
	timeout = 'timeout' in exp.info._exp_yml and runtime > float(exp.info._exp_yml['timeout'])

	# Create the status file to signal that we are finished.
	status_dict = {'timeout': timeout, 'walltime': runtime,
			'status': status, 'signal': sigcode}
	with open(run.output_file_path('status.tmp'), "w") as f:
		yaml.dump(status_dict, f)
	os.rename(run.output_file_path('status.tmp'), run.output_file_path('status'))





import errno
import os
import re
import shutil
import yaml

def expand_at_params(s, fn, listfn=None):
	def subfn(m):
		result = fn(m.group(1))
		if result is None:
			raise RuntimeError("Unexpected @-parameter {}".format(p))
		return result

	if isinstance(s, list):
		seq = [ ]
		for item in s:
			# Try to replace the whole item.
			if listfn is not None:
				m = re.fullmatch(r'@(\w+)@', item)
				if m:
					result = listfn(m.group(1))
					if result is not None:
						seq.extend(result)
						continue

			seq.append(re.sub(r'@(\w+)@', subfn, item))
		return seq
	else:
		assert isinstance(s, str)
		return re.sub(r'@(\w+)@', subfn, s)

def try_mkdir(path):
	try:
		os.mkdir(path)
	except OSError as error:
		if error.errno != errno.EEXIST:
			raise

def try_rmfile(path):
	try:
		os.remove(path)
	except OSError as error:
		if error.errno != errno.EEXIST:
			raise

def try_rmtree(path):
	try:
		shutil.rmtree(path)
	except OSError as error:
		if error.errno != errno.ENOENT:
			raise

def touch(path):
	with open(path, 'w') as f:
		pass

def read_setup_file(setup_file):
	with open(setup_file, 'r') as f:
		setup_dict = yaml.load(f, Loader=yaml.BaseLoader)
	return setup_dict

def validate_setup_file(setup_file):
	""" Reads, validates and sanitizes the setup file
	"""

	setup_dict = read_setup_file(setup_file)

	if 'instdir' not in setup_dict:
		setup_dict['instdir'] = './instances';

	return setup_dict

def compute_network_size(path, out):
	import networkit as nk
	try:
		g = nk.readGraph(path, nk.Format.EdgeList,
				separator=' ', firstNode=0, continuous=False, directed=False)
	except Exception: # Exception due the attempt of reading a non-network file
		return
	data = {
		'n': g.numberOfNodes(),
		'm': g.numberOfEdges()
	}
	yaml.dump(data, out, default_flow_style=False)





from enum import Enum, IntEnum
import os.path
import subprocess

from . import util

def make_builds(cfg, revision, infos):
	order = compute_order(cfg, infos)

	print("simexpal: Making builds {} @ {}".format(', '.join([info.name for info in order]),
			revision.name))
	for info in order:
		make_build_in_order(cfg, cfg.get_build(info.name, revision))

def compute_order(cfg, desired):
	class State(Enum):
		NULL = 0
		EXPANDING = 1
		ORDERED = 2

	order = [ ]

	# The following code does a topologic sort of the desired items.
	# It lazily expands their dependencies using expand().
	stack = [ ]
	state = dict()

	def expand(info):
		for name in info.requirements:
			yield cfg.get_build_info(name)

	def visit(info):
		if info.name not in state:
			links = expand(info)
			state[info.name] = State.EXPANDING
			stack.append((info, list(links)))
		elif state[info.name] == State.EXPANDING:
			for (circ_action, circ_subject, _) in stack:
				print(Action.strings[circ_action], circ_subject.name)
			raise RuntimeError("Program has circular dependencies")
		else:
			# Programs that are already ordered do not need to be considered again.
			assert state[info.name] == State.ORDERED

	for info in desired:
		visit(info)

		while stack:
			(info, rem_links) = stack[-1]
			if not rem_links:
				assert state[info.name] == State.EXPANDING
				state[info.name] = State.ORDERED
				stack.pop()
				order.append(info)
			else:
				visit(rem_links.pop())

	return order

def make_build_in_order(cfg, build):
	util.try_mkdir('builds/')

	def substitute(var):
		if var == 'THIS_CLONE_DIR':
			return build.clone_dir
		elif var == 'THIS_PREFIX_DIR':
			return build.prefix_dir
		elif var == 'PARALLELISM':
			nthreads = len(os.sched_getaffinity(0))
			return str(nthreads)

	# Build the environment.
	def prepend_env(var, items):
		if(var in os.environ):
			return ':'.join(items) + ':' + os.environ[var]
		return ':'.join(items)

	recursive_infos = build.info.traverse_requirements()
	recursive_builds = [cfg.get_build(info.name, build.revision) for info in recursive_infos]

	def collect_prefix_paths(subdir):
		return [os.path.join(req.prefix_dir, subdir) for req in recursive_builds]
	pkgconfig_paths = collect_prefix_paths('lib/pkgconfig')

	base_environ = os.environ.copy()
	base_environ['PKG_CONFIG_PATH'] = prepend_env('PKG_CONFIG_PATH', pkgconfig_paths)

	# Determine which phases to run.
	class Phase(IntEnum):
		NULL = 0
		CHECKOUT = 1
		REGENERATE = 2
		CONFIGURE = 3
		COMPILE = 4
		INSTALL = 5

	done_phases = set()
	if os.access(os.path.join(build.prefix_dir, 'installed.simexpal'), os.F_OK):
		done_phases.add(Phase.INSTALL)
	if os.access(os.path.join(build.compile_dir, 'compiled.simexpal'), os.F_OK):
		done_phases.add(Phase.COMPILE)
	if os.access(os.path.join(build.compile_dir, 'configured.simexpal'), os.F_OK):
		done_phases.add(Phase.CONFIGURE)
	if os.access(os.path.join(build.clone_dir, 'regenerated.simexpal'), os.F_OK):
		done_phases.add(Phase.REGENERATE)
	if os.access(os.path.join(build.clone_dir, 'checkedout.simexpal'), os.F_OK):
		done_phases.add(Phase.CHECKOUT)

	def want_phase(phase):
		# TODO: Support additional phase section modes. For example:
		#       - Clean rebuilds from scratch. This should be prefered for production use.
		#       - Running individual phases. This should help with debugging.
		return not done_phases or phase > max(done_phases)

	# Perform the actual build phases.
	def log_phase(step):
		print("simexpal: Running {}-phase for build {}".format(step, build.name))

	did_work = False

	def do_step(step_yml, workdir=None):
		environ = base_environ.copy()
		if 'environ' in step_yml:
			for (var, value) in step_yml['environ'].items():
				environ[var] = util.expand_at_params(value, substitute)
		args = [util.expand_at_params(arg, substitute) for arg in step_yml['args']]
		subprocess.check_call(args, cwd=workdir, env=environ)

	if want_phase(Phase.CHECKOUT):
		log_phase('checkout')

		# Recreate the source directory.
		util.try_rmtree(build.clone_dir)

		subprocess.check_call(['git', 'clone', build.info._build_yml['git'], build.clone_dir])
		subprocess.check_call(['git', '--git-dir', build.clone_dir + '/.git',
				'--work-tree', build.clone_dir,
				'checkout',
				build.revision.version_for_build(build.name)])
		util.touch(os.path.join(build.clone_dir, 'checkedout.simexpal'))
		did_work = True

	if want_phase(Phase.REGENERATE):
		log_phase('regenerate')
		if 'regenerate' in build.info._build_yml:
			for step_yml in build.info._build_yml['regenerate']:
				do_step(step_yml, build.clone_dir)
		util.touch(os.path.join(build.clone_dir, 'regenerated.simexpal'))
		did_work = True

	if want_phase(Phase.CONFIGURE):
		log_phase('configure')

		# Recreate the compilation directory.
		util.try_rmtree(build.compile_dir)
		util.try_mkdir(build.compile_dir)

		if 'configure' in build.info._build_yml:
			for step_yml in build.info._build_yml['configure']:
				do_step(step_yml, workdir=build.compile_dir)
		util.touch(os.path.join(build.compile_dir, 'configured.simexpal'))
		did_work = True

	if want_phase(Phase.COMPILE):
		log_phase('compile')
		if 'compile' in build.info._build_yml:
			for step_yml in build.info._build_yml['compile']:
				do_step(step_yml, workdir=build.compile_dir)
		util.touch(os.path.join(build.compile_dir, 'compiled.simexpal'))
		did_work = True

	if want_phase(Phase.INSTALL):
		log_phase('install')

		# Recreate the prefix directory.
		util.try_rmtree(build.prefix_dir)
		util.try_mkdir(build.prefix_dir)

		if 'install' in build.info._build_yml:
			for step_yml in build.info._build_yml['install']:
				do_step(step_yml, workdir=build.compile_dir)
		util.touch(os.path.join(build.prefix_dir, 'installed.simexpal'))
		did_work = True

	if not did_work:
		print("simexpal: Nothing to do for {}".format(build.name))





import json
import os
import socket

from . import common

class QueueLauncher(common.Launcher):
	def submit(self, config, run):
		if not common.lock_run(run):
			return
		common.create_run_file(run)

		print("Launching experiment '{}', instance '{}' on local machine".format(
				run.experiment.name, run.instance.filename))

		m = {
			'basedir': config.basedir,
			'experiment': run.experiment.name,
			'instance': run.instance.filename,
			'repetition': run.repetition
		}

		sockpath = os.path.expanduser('~/.extlq.sock')
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		s.connect(sockpath)
		s.send((json.dumps(m) + '\n').encode())
		s.close()



#!/usr/bin/env python3

from setuptools import setup

with open('README.md', 'r') as f:
	readme = f.read()

setup(
	# First, state all metadata about the package.
	name='simexpal',
	version='0.2',
	description='Tool to Simplify Experimental Algorithmics',
	url='https://github.com/hu-macsy/simexpal',
	author='Alexander van der Grinten, MACSy group HU-Berlin',
	author_email='alexander.vandergrinten@gmail.com',
	license='MIT',
	long_description=readme,
	long_description_content_type='text/markdown',

	# Now, set the actual Python configuration.
	packages=['simexpal', 'simexpal.launch'],
	scripts=['scripts/simex'],
	install_requires=[
		'argcomplete',
		'requests',
		'pyyaml'
	]
)





from . import common

class ForkLauncher(common.Launcher):
	def submit(self, config, run):
		if not common.lock_run(run):
			return
		common.create_run_file(run)

		print("Launching experiment '{}', instance '{}' on local machine".format(
				run.experiment.name, run.instance.filename))
		common.invoke_run(run)

